# Copyright 2013-2018 Rumma & Ko Ltd
# License: BSD, see LICENSE for more details.

# This module has no docstring because it is to be execfile'd
# from `setup.py`, `etgen/__init__.py` and possibly some external
# tools, too.

install_requires = ['six', 'django', 'future', 'lxml', 'atelier']

SETUP_INFO = dict(
    name='etgen',
    version='0.0.3',
    install_requires=install_requires,
    description="Use ElementTree to generate html, rst and other markup",
    license='Free BSD',
    test_suite='tests',
    author='Luc Saffre',
    author_email='luc@lino-framework.org',
    url="http://etgen.lino-framework.org",
    long_description="""\

Utilities for generating html, xml and rst content from an
`ElementTree
<https://docs.python.org/2/library/xml.etree.elementtree.html>`_.

Inspired by Frederik Lundh's `ElementTree Builder
<http://effbot.org/zone/element-builder.htm>`_.

The central project homepage is http://etgen.lino-framework.org

""",
    classifiers="""\
Programming Language :: Python
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Framework :: Sphinx :: Extension
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent""".splitlines())

SETUP_INFO.update(packages=[n for n in """
etgen
etgen.intervat
etgen.odf
etgen.sepa
""".splitlines() if n])

SETUP_INFO.update(package_data=dict(), include_package_data=True)


def add_package_data(package, *patterns):
    l = SETUP_INFO['package_data'].setdefault(package, [])
    l.extend(patterns)
    return l


add_package_data('etgen', 'etgen/sepa/XSD/*.xsd')

# print(20180216, SETUP_INFO['package_data'])
# raw_input()



# -*- coding: UTF-8 -*-
# Copyright 2013-2018 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""
Convert an :mod:`etgen.html` element to a reStructuredText string.

If `stripped` is `True`, output will be more concise and optimized
for console output, but possibly not valid reStructuredText.

Usage examples:

>>> from etgen.html import E
>>> e = E.p("This is a ", E.b("first"), " test.")
>>> print(html2rst(e, True))
This is a **first** test.
<BLANKLINE>

>>> e = E.p(E.b("This")," is another test.")
>>> print(html2rst(e, True))
**This** is another test.
<BLANKLINE>

>>> e = E.p(E.b("This")," is ",E.em("another")," test.")
>>> print(html2rst(e, True))
**This** is *another* test.
<BLANKLINE>

>>> url = "http://example.com"
>>> e = E.p(E.b("This")," is ",E.a("a link",href=url),".")
>>> print(html2rst(e, True))
**This** is `a link <http://example.com>`__.
<BLANKLINE>

>>> e = E.p("An empty bold text:",E.b(""))
>>> print(html2rst(e, True))
An empty bold text:
<BLANKLINE>

>>> e = E.ul(E.li("First"), E.li("Second"))
>>> print(html2rst(e, True))
<BLANKLINE>
First
Second
<BLANKLINE>

>>> e = E.h1("A header")
>>> print(html2rst(e, True))
========
A header
========
<BLANKLINE>

For images we render the ``alt`` text between brackets:

>>> e = E.img(src="http://example.com/images/1.jpg", alt="1")
>>> print(html2rst(e, True))
[img 1]

If there is no ``alt`` text, render the content of ``src``:

>>> e = E.img(src="http://example.com/images/1.jpg")
>>> print(html2rst(e, True))
[img http://example.com/images/1.jpg]
"""

from __future__ import unicode_literals

from atelier import rstgen

from etgen import etree

NEWLINE_TAGS = set(['p', 'thead', 'tr', 'li'])
IGNORED_TAGS = set(['tbody', 'table', 'div', 'span', 'br', 'ul', 'ol'])


def html2rst(e, stripped=False):
    """
    Convert an element tree to reStructuredText.
    """
    #~ print("20120613 html2odftext()", e.tag, e.text)
    rst = ''
    if e.tag in ('p', 'li'):
        if not stripped:
            rst += '\n\n'

    elif e.tag in ('ul', 'ol'):
        rst += '\n'
    elif e.tag == 'br':
        if stripped:
            rst += '\n'
        else:
            rst += ' |br| \n'
    elif e.tag == 'b':
        rst += '**'
    elif e.tag == 'em' or e.tag == 'i':
        rst += '*'
    elif e.tag == 'a':
        rst += '`'

    if e.text:
        rst += e.text
    for child in e:
        rst += html2rst(child, stripped)

    if e.tag in NEWLINE_TAGS:
        if stripped:
            rst += '\n'
        else:
            rst += '\n\n'
    elif e.tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
        rst = rstgen.header(int(e.tag[1]), rst.strip()).strip()
        if stripped:
            rst += '\n'
        else:
            rst = '\n\n' + rst + '\n\n'
    elif e.tag == 'b':
        if rst == '**':
            rst = ''
        else:
            rst += '**'
    elif e.tag == 'em' or e.tag == 'i':
        if rst == '*':
            rst = ''
        else:
            rst += '*'
    elif e.tag == 'a':
        rst += ' <%s>`__' % e.get('href')
    elif e.tag == 'img':
        text = e.get('alt') or e.get('src')
        rst += '[img %s]' % text
    elif e.tag in ('td', 'th'):
        rst += ' '
    else:
        if e.tag not in IGNORED_TAGS:
            raise Exception("20150723 %s" % e.tag)

    if e.tail:
        rst += e.tail

    return rst


# def html2rst(e):
#     return _html2rst(e).strip()


class RstTable(rstgen.Table):

    """
    A table containing elementtree HTML:

.. complextable::
    :header:

    Code <NEXTCELL> Result <NEXTROW>

    >>> from etgen.html import E
    >>> headers = [E.p("A ", E.b("formatted"), " header"), "A plain header"]
    >>> rows = [[1,2], [3,4]]
    >>> print(RstTable(headers).to_rst(rows))
    ======================== ================
     A **formatted** header   A plain header
    ------------------------ ----------------
     1                        2
     3                        4
    ======================== ================
    <BLANKLINE>

    <NEXTCELL>

    ======================== ================
     A **formatted** header   A plain header
    ------------------------ ----------------
     1                        2
     3                        4
    ======================== ================
    """

    def format_value(self, v):
        if etree.iselement(v):
            return html2rst(v, True).strip()
        return super(RstTable, self).format_value(v).strip()





# Copyright 2011-2018 Rumma & Ko Ltd
# License: BSD, see LICENSE for more details.
"""
This is the :mod:`etgen` package.

Inspired by Frederik Lundh's
`ElementTree Builder
<http://effbot.org/zone/element-builder.htm>`_

.. autosummary::
   :toctree:

   html
   html2rst
   intervat
   odf
   sepa
   utils


>>> E = Namespace('http://my.ns',
...    "bar baz bar-o-baz foo-bar class def".split())

>>> bob = E.bar_o_baz()
>>> baz = E.add_child(bob, 'baz', class_='first')
>>> print E.tostring(baz)
<baz xmlns="http://my.ns" class="first" />

>>> bob = E.bar_o_baz('Hello', class_='first', foo_bar="3")
>>> print E.tostring(bob)
<bar-o-baz xmlns="http://my.ns" class="first" foo-bar="3">Hello</bar-o-baz>

The following reproduces a pifall. Here is the initial code:

>>> E = Namespace(None, "div br".split())
>>> bob = E.div("a", E.br(), "b", E. br(), "c", E.br(), "d")
>>> print E.tostring(bob)
<div>a<br />b<br />c<br />d</div>

The idea is to use :func:`join_elems` to insert the <br> tags:

>>> from etgen.utils import join_elems

But surprise:

>>> elems = join_elems(["a", "b", "c", "d"], sep=E.br())
>>> print E.tostring(E.div(*elems))
<div>a<br />bcd<br />bcd<br />bcd</div>

What happened here is that the same `<br>` element instance was being
inserted multiple times at different places.  The correct usage is
without the parentheses so that `join_elems` instantiates each time a
new element:

>>> elems = join_elems(["a", "b", "c", "d"], sep=E.br)
>>> print E.tostring(E.div(*elems))
<div>a<br />b<br />c<br />d</div>



"""

import os

from .setup_info import SETUP_INFO

fn = os.path.join(os.path.dirname(__file__), 'setup_info.py')
exec(compile(open(fn, "rb").read(), fn, 'exec'))

__version__ = SETUP_INFO['version']

intersphinx_urls = dict(docs="http://etgen.lino-framework.org")
srcref_url = 'https://github.com/lino-framework/etgen/blob/master/%s'




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2013-2018 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""
"""

import sys
from lxml import etree


XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema"
XSD = "{%s}" % XSD_NAMESPACE

NSMAP = dict(xsd=XSD_NAMESPACE)


def xsdnames(fn, nsname):
    tree = etree.parse(fn)
    root = tree.getroot()
    names = set()

    def collect(e):
        na = e.get('name', None)
        if na is not None:
            names.add(na)
        for child in e:
            collect(child)
    collect(root)
    return ' '.join(names)


def main(fn):
    print xsdnames(fn, 'ns0')

if __name__ == "__main__":
    main(sys.argv[1])



# Copyright 2011-2018 Rumma & Ko Ltd
# License: BSD, see LICENSE for more details.
"""
Some utility functions.
"""

from builtins import str
# from builtins import object
from future.types import newstr
import six
import datetime
from functools import partial
from etgen import etree
from django.utils.functional import Promise
from django.utils.encoding import force_text


def join_elems(elems, sep=' '):
    """
    Examples:

    >>> join_elems([1, 2, 3])
    [1, ' ', 2, ' ', 3]
    >>> join_elems([1, 2, 3],' / ')
    [1, ' / ', 2, ' / ', 3]
    >>> join_elems([])
    []

    """
    if not callable(sep):
        sep_value = sep

        def sep():
            return sep_value
    l = []
    s = None
    for e in elems:
        if s is not None:
            l.append(s)
        s = sep()
        l.append(e)
    return l



def pretty_print(elem):
    """
    Return a pretty-printed XML string for the Element.
    """
    return prettify(etree.tostring(elem, 'utf-8'))
    # the following also indented:
    # from http://renesd.blogspot.com/2007/05/pretty-print-xml-with-python.html
    # via http://broadcast.oreilly.com/2010/03/pymotw-creating-xml-documents.html
    #~ from xml.dom import minidom
    #~ rough_string = etree.tostring(elem, 'utf-8')
    #~ reparsed = minidom.parseString(rough_string)
    #~ return reparsed.toprettyxml(indent="  ")


def prettify(s):
    return s.replace('><', '>\n<')


def compatstr(s):
    """
    The `python-future <http://python-future.org/>`__ package
    introduces a special helper class `newstr` which simulates, under
    Python 2, the behaviour of Python 3 strings.  But
    `xml.etree.ElementTree
    <https://docs.python.org/2/library/xml.etree.elementtree.html>`__
    in Python 2 doesn't know about `python-future` and produces
    invalid XML when you feed it with such a string.

    So this function converts any `newstr` back to a real newstr.

    TODO: Not yet tested under Python 3. At the best it is just
    unefficient.
    """
    # assert_pure(s)
    if six.PY2 and isinstance(s, newstr):
        return six.text_type(s)
    return s

RESERVED_WORDS = frozenset("""
and       del       from      not       while
as        elif      global    or        with
assert    else      if        pass      yield
break     except    import    print
class     exec      in        raise
continue  finally   is        return
def       for       lambda    try
""".split())

TYPEMAP = {
    #~ datetime.datetime: py2str,
    #~ IncompleteDate : lambda e,v : str(v),
    datetime.datetime: lambda e, v: v.strftime("%Y%m%dT%H%M%S"),
    datetime.date: lambda e, v: v.strftime("%Y-%m-%d"),
    int: lambda e, v: str(v),
}


class Namespace(object):
    """
    An XML namespace.  Base class for :class:`etgen.html.HtmlNamespace`
    and the namespaces defined in :mod:`etgen.intervat`.
    """
    prefix = None
    targetNamespace = None
    names = None

    def __init__(self, targetNamespace=None, names=None, prefix=None):
        #~ if prefix is not None:
            #~ self.prefix = prefix
        #~ kw.setdefault('typemap',TYPEMAP)
        #~ kw.setdefault('makeelement',self.makeelement)
        #~ nsmap = kw.setdefault('nsmap',{})

        if prefix is not None:
            self.prefix = prefix
        if names is not None:
            self.names = names
        if targetNamespace is not None:
            self.targetNamespace = targetNamespace
        if self.targetNamespace is not None:
            #~ kw.update(namespace=self.targetNamespace)

            self._ns = '{' + self.targetNamespace + '}'
            if self.prefix is not None:
                etree.register_namespace(self.prefix, self.targetNamespace)
            #~ if prefix:
            #~ nsmap[prefix] = self.targetNamespace
        #~ if used_namespaces is not None:
            #~ self.used_namespaces = used_namespaces
        #~ if self.used_namespaces is not None:
            #~ for ns in self.used_namespaces:
                #~ nsmap[ns.prefix] = ns.targetNamespace
        #~ self._element_maker = ElementMaker(**kw)
        #~ self._source_elements = {}
        if self.names is not None:
            self.define_names(self.names)
        self.setup_namespace()

    def iselement(self, *args, **kw):
        return etree.iselement(*args, **kw)

    def setup_namespace(self):
        pass

    def tostring(self, element, *args, **kw):
        class dummy(object):
            pass
        data = []
        file = dummy()
        file.write = data.append
        if self.targetNamespace is not None:
            kw.setdefault('default_namespace', self.targetNamespace)
        etree.ElementTree(element).write(file, *args, **kw)
        return b"".join(data).decode("utf-8")

    def tostring_pretty(self, *args, **kw):
        #~ kw.setdefault('xml_declaration',False)
        #~ kw.setdefault('encoding','utf-8')
        #~ kw.update(xml_declaration=False)
        #~ kw.update(encoding='utf-8')
        s = self.tostring(*args, **kw)
        #~ return s
        #~ return minidom.parseString(s).toprettyxml(indent="  ")
        return prettify(s)

    def addns(self, tag):
        if self.targetNamespace is None or tag[0] == "{":
            return tag
        return self._ns + tag

    def makeattribs(self, **kw):
        #~ ns = self._element_maker._namespace
        #~ if ns is None: return kw
        xkw = dict()
        for k, v in list(kw.items()):
            k = getattr(self, k).args[0]  # convert iname to tagname
            xkw[self.addns(compatstr(k))] = compatstr(v)
        return xkw

    def create_element(self, tag, *children, **attrib):
        nsattrib = self.makeattribs(**attrib)
        tag = self.addns(tag)
        elem = etree.Element(tag, nsattrib)
        for item in children:
            if isinstance(item, Promise):
                item = force_text(item)
                # assert_pure(item)
            if isinstance(item, dict):
                elem.attrib.update(self.makeattribs(**item))
            elif isinstance(item, six.string_types):
                # assert_pure(item)
                #~ if len(elem) and len(elem[-1]) == 0:
                if len(elem):
                    last = elem[-1]
                    last.tail = (last.tail or "") + item
                else:
                    elem.text = (elem.text or "") + item
            elif etree.iselement(item):
                elem.append(item)
            else:
                raise TypeError("bad argument: %r" % item)
            #~ print "20130805 added %s --> %s" % (item,self.tostring(elem))
        return elem

    def define_names(self, names):
        if isinstance(names, six.string_types):
            raise Exception("{} must now call itself split().".format(
                self))
        for tag in names:
            iname = tag.replace("-", "_")
            iname = iname.replace(".", "_")
            #~ if iname in ('class','for','in','def'):
            if iname in RESERVED_WORDS:
                iname += "_"
            #~ setattr(self,iname,getattr(self._element_maker,name))
            p = partial(self.create_element, tag)
            setattr(self, iname, p)

    def getnsattr(self, elem, name):
        #~ if self.targetNamespace is None or name.startswith('{'):
            #~ return elem.get(name)
        return elem.get(self._element_maker._namespace + name)

    #~ def update_attribs(self,root,**kw):
    def update(self, root, **kw):
        root.attrib.update(self.makeattribs(**kw))

    def add_child(self, parent, _name, *args, **kw):
        ecl = getattr(self, _name)
        #~ kw = self.makeattribs(**kw)
        #~ print 20120420, kw
        e = ecl(*args, **kw)
        parent.append(e)
        return e

    def fromstring(self, s, **kwargs):
        """Build an element tree from the given XML source string.

        This just forwards to the
        :meth:`xml.etree.ElementTree.fromstring` library function.
        See the `Parsing XML
        <https://docs.python.org/2.7/library/xml.etree.elementtree.html#parsing-xml>`__
        section of the Python docs.

        """
        return etree.etree.fromstringlist([s], **kwargs)

    def raw(self, *args):
        """Parses the given string into an XML Element."""
        return RAW(*args)

RAW = etree.XML


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



from atelier.invlib.ns import ns
ns.setup_from_tasks(
    globals(), "etgen",
    blogref_url="http://luc.lino-framework.org",
    revision_control_system='git',
    # tolerate_sphinx_warnings=True,
    cleanable_files=['docs/api/etgen.*'])




# -*- coding: UTF-8 -*-
# doctest etgen/html
# Adapted copy from lxml\src\lxml\html\builder.py
# --------------------------------------------------------------------
# The ElementTree toolkit is
# Copyright (c) 1999-2004 by Fredrik Lundh
# Modifications in this file are
# Copyright (c) 2012-2018 Rumma & Ko Ltd
# --------------------------------------------------------------------

"""Defines an ElementTree Builder for generating HTML documents.

.. data:: E

    The global ElementTree Builder object.

Usage:

>>> from etgen.html import E
>>> html = E.html(
...            E.head( E.title("Hello World") ),
...            E.body(
...              E.h1("Hello World !"),
...              class_="main"
...            )
...        )

>>> print (E.tostring_pretty(html))
<html>
<head>
<title>Hello World</title>
</head>
<body class="main">
<h1>Hello World !</h1>
</body>
</html>


>>> kw = dict(title=u'Ein süßes Beispiel')
>>> kw.update(href="foo/bar.html")
>>> btn = E.button(type='button', class_='x-btn-text x-tbar-upload')
>>> html = E.a(btn, **kw)
>>> print (E.tostring_pretty(html))
<a href="foo/bar.html" title="Ein s&#252;&#223;es Beispiel">
<button class="x-btn-text x-tbar-upload" type="button" />
</a>

You can also do the opposite, i.e. parse HTML:

>>> html = E.raw('''<a href="foo/bar.html"
... title="Ein s&#252;&#223;es Beispiel">
... <button class="x-btn-text x-tbar-upload" type="button" />
... </a>''')
>>> print(E.tostring_pretty(html))
<a href="foo/bar.html" title="Ein s&#252;&#223;es Beispiel">
<button class="x-btn-text x-tbar-upload" type="button" />
</a>


>>> print(E.tostring(E.raw(
...     '<ul type="disc"><li>First</li><li>Second</li></ul>')))
<ul type="disc"><li>First</li><li>Second</li></ul>

>>> html = E.div(E.p("First"), E.p("Second"))
>>> print(E.tostring_pretty(html))
<div>
<p>First</p>
<p>Second</p>
</div>
>>> html.attrib['class'] = 'htmlText'
>>> print(E.tostring_pretty(html))
<div class="htmlText">
<p>First</p>
<p>Second</p>
</div>


"""

from __future__ import unicode_literals
from builtins import str
from builtins import object

import types
from xml.etree import ElementTree as ET
from lxml.etree import HTML

from etgen.utils import join_elems
from etgen.utils import Namespace
from etgen.html2rst import html2rst
# from htmlentitydefs import name2codepoint

# ENTITIES = {}
# ENTITIES.update((x, unichr(i)) for x, i in name2codepoint.iteritems())


# def CreateParser():
#     """Every string that is being parsed must get its own parser instance.
#     This is because "Due to limitations in the Expat library used by
#     pyexpat, the xmlparser instance returned can only be used to parse
#     a single XML document. Call ParserCreate for each document to
#     provide unique parser instances. (`docs.python.org
#     <https://docs.python.org/2/library/pyexpat.html>`_)

#     """
#     p = ET.XMLParser()
#     # PARSER.entity.update(htmlentitydefs.entitydefs)
#     p.entity = ENTITIES
#     assert 'otilde' in p.entity
#     assert 'eacute' in p.entity
#     assert 'nbsp' in p.entity
#     return p

# class RAW_HTML_STRING(unicode):
#     pass


class HtmlNamespace(Namespace):
    """The HTML namespace.
    This is instantiated as ``E``.
    """

    def tostring(self, v, *args, **kw):
        # if isinstance(v, types.GeneratorType):
        if isinstance(v, (types.GeneratorType, list, tuple)):
            return "".join([self.tostring(x, *args, **kw) for x in v])
        if self.iselement(v):
            # kw.setdefault('method', 'html')
            return super(HtmlNamespace, self).tostring(v, *args, **kw)
        return str(v)

    def to_rst(self, v, stripped=True):
        if isinstance(v, types.GeneratorType):
            return "".join([self.to_rst(x, stripped) for x in v])
        if self.iselement(v):
            return html2rst(v, stripped)
        return str(v)

    # def raw(self, raw_html):
    #     return RAW_HTML_STRING(raw_html)

    def raw(self, raw_html):
        """
        Parses the given string into an HTML Element.

        It the string contains a a single top-level element, then this
        element is returned. Otherwise return the wrapping ``body``
        element.
        """
        # print 20151008, raw_html

        # the lxml parser wraps `<html><body>...</body></html>` around
        # the snippet, but we don't want it.
        try:
            root = HTML(raw_html)[0]
            # root = E.html(raw_html)[0]
        except Exception as e:
            return E.p("Invalid HTML ({}) in {}".format(e, raw_html))
        if len(root) == 1:
            return root[0]
        return root
        # try:
        #     return self.fromstring(raw_html, parser=CreateParser())
        # except ET.ParseError as e:
        #     raise Exception("ParseError {0} in {1}".format(e, raw_html))


E = HtmlNamespace(None, set("""
a
abbr
acronym
address
alt
applet
area
b
base
basefont
bdo
big
blockquote
body
br
button
caption
center
cite     
code     
col      
colgroup 
dd       
del      
dfn      
dir      
div      
dl       
dt       
em       
fieldset 
font     
form     
frame    
frameset 
h1     
h2     
h3     
h4     
h5
h6
head
height
hr     
html   
i      
iframe 
img    
input  
ins    
isindex 
kbd 
label 
legend 
li 
link 
map 
menu 
meta 
name
noframes 
noscript 
object 
ol 
optgroup 
option 
p
param 
pre 
q 
s 
samp
script
select
small
span
strike
strong
style
sub
sup
table
tbody
td
textarea
tfoot
th
thead
title
tr
tt
u
ul
var

class
id
bgcolor
cellspacing
width
align
valign
href
type
rel
target
value
onclick
src
rows
data-toggle
tabindex
placeholder
""".split()))


def table_header_row(*headers, **kw):
    return E.tr(*[E.th(h, **kw) for h in headers])


def table_body_row(*cells, **kw):
    return E.tr(*[E.td(h, **kw) for h in cells])


class Table(object):
    """A pythonic representation of a ``<table>`` with ``<head>``,
    ``<foot>`` and ``<body>``.

    """
    def __init__(self):
        self.clear()

    def clear(self):
        self.head = []
        self.foot = []
        self.body = []
        self.attrib = dict()

    def add_header_row(self, *args, **kw):
        e = table_header_row(*args, **kw)
        self.head.append(e)
        return e

    def add_footer_row(self, *args, **kw):
        e = table_body_row(*args, **kw)
        self.foot.append(e)
        return e

    def add_body_row(self, *args, **kw):
        e = table_body_row(*args, **kw)
        self.body.append(e)
        return e

    def as_element(self):
        children = []
        if self.head:
            children.append(E.thead(*self.head))
        if self.foot:
            children.append(E.tfoot(*self.foot))
        if self.body:
            children.append(E.tbody(*self.body))
        return E.table(*children, **self.attrib)


class Document(object):
    """A pythonic representation of a ``<body>`` with a ``<title>`` and
    some ``<head>`` tags for stylesheets.

    """

    def __init__(self, title, stylesheets=None):
        self.title = title
        self.body = []
        self.stylesheets = stylesheets or []

    def add_stylesheet(self, url):
        self.stylesheets.append(url)

    def add_table(self):
        t = Table()
        self.body.append(t)
        return t

    def write(self, *args, **kw):
        ET.ElementTree(self.as_element()).write(*args, **kw)

    def as_element(self):
        body = []
        for e in self.body:
            if isinstance(e, Table):
                body.append(e.as_element())
            else:
                body.append(e)
        headers = []
        for css in self.stylesheets:
            headers.append(E.link(rel="stylesheet", type="text/css", href=css))
        headers.append(E.title(self.title))

        return E.html(
            E.head(*headers),
            E.body(*body)
        )


def lines2p(lines, min_height=0, **attrs):
    """Convert the given list of text lines `lines` into a paragraph
(``<p>``) with one ``<br>`` between each line. If optional
`min_height` is given, add empty lines if necessary.

    Examples:

    >>> print(E.tostring(lines2p(['first', 'second'])))
    <p>first<br />second</p>

    >>> print(E.tostring(lines2p(['first', 'second'], min_height=5)))
    <p>first<br />second<br /><br /><br /></p>

    If `min_height` is specified, and `lines` contains more items,
    then we don't truncate:

    >>> print(E.tostring(lines2p(['a', 'b', 'c', 'd', 'e'], min_height=4)))
    <p>a<br />b<br />c<br />d<br />e</p>

    This also works:

    >>> print(E.tostring(lines2p([], min_height=5)))
    <p><br /><br /><br /><br /></p>

    """
    while len(lines) < min_height:
        lines.append('')
    lines = join_elems(lines, E.br)
    return E.p(*lines, **attrs)


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()
